@@ -1,6 +1,6 @@
 ;;; projectile-test.el
 
-;; Copyright © 2019 Bozhidar Batsov
+;; Copyright © 2011-2020 Bozhidar Batsov
 
 ;; Author: Bozhidar Batsov <bozhidar@batsov.com>
 
@@ -28,6 +28,7 @@
 (require 'projectile)
 (require 'buttercup)
 
+
 (message "Running tests on Emacs %s" emacs-version)
 
 ;; TODO: Revise this init logic
@@ -65,9 +66,29 @@ You'd normally combine this with `projectile-test-with-sandbox'."
                files)
      ,@body))
 
+(defmacro projectile-test-with-files-using-custom-project (files project-options &rest body)
+  "Evaluate BODY with the custom project having PROJECT-OPTIONS with FILES."
+  (declare (indent 2) (debug (sexp sexp &rest form)))
+  `(let ((projectile-indexing-method 'native)
+         (projectile-projects-cache (make-hash-table :test 'equal))
+         (projectile-projects-cache-time (make-hash-table :test 'equal))
+         (projectile-enable-caching t))
+     ,@(mapcar (lambda (file)
+                 (let* ((path (concat "project/" file))
+                        (dir (file-name-directory path)))
+                   (if (string-suffix-p "/" file)
+                       `(make-directory ,path t)
+                     `(progn
+                        (make-directory ,dir t)
+                        (with-temp-file ,path)))))
+               files)
+     (projectile-register-project-type 'sample-project '("somefile") ,@project-options)
+     (spy-on 'projectile-project-type :and-return-value 'sample-project)
+     (spy-on 'projectile-project-root :and-return-value (file-truename (expand-file-name "project/")))
+     ,@body))
+
 (defun projectile-test-tmp-file-path ()
-  "Return a filename suitable to save data to in the
-test temp directory"
+  "Return a filename suitable to save data to in the test temp directory."
   (concat projectile-test-path
           "/tmp/temporary-file-" (format "%d" (random))
           ".eld"))
@@ -215,7 +236,7 @@ test temp directory"
     (spy-on 'projectile-project-vcs :and-return-value 'git)
     (spy-on 'projectile-get-repo-ignored-files :and-return-value '("path/unignored-file"))
     (spy-on 'projectile-get-repo-ignored-directory :and-call-fake
-            (lambda (project vcs dir)
+            (lambda (project dir vcs)
               (list (concat dir "unignored-file"))))
     (let ((projectile-globally-unignored-directories '("path")))
       (expect (projectile-add-unignored nil nil '("file")) :to-equal '("file" "path/unignored-file"))
@@ -623,6 +644,86 @@ test temp directory"
         (expect (projectile-project-root) :to-equal correct-project-root))))))
 
 (describe "projectile-grep"
+  (describe "multi-root grep"
+    (after-each
+      (cl-flet ((grep-buffer-p (b) (string-prefix-p "*grep" (buffer-name b))))
+        (let ((grep-buffers (cl-remove-if-not #'grep-buffer-p (buffer-list))))
+          (dolist (grep-buffer grep-buffers)
+            (let ((kill-buffer-query-functions nil))
+              (kill-buffer grep-buffer))))))
+    (it "grep multi-root projects"
+      (projectile-test-with-sandbox
+        (projectile-test-with-files
+            ("project/bar/"
+             "project/baz/")
+          (cd "project")
+          (with-temp-file ".projectile" (insert (concat "+/baz\n"
+                                                        "+/bar\n")))
+          (with-temp-file "foo.txt" (insert "hi"))
+          (with-temp-file "bar/bar.txt" (insert "hi"))
+          (with-temp-file "baz/baz.txt" (insert "hi"))
+          (with-current-buffer (find-file-noselect ".projectile" t)
+            (let ((grep-find-template "<X>")
+                  grep-find-ignored-directories grep-find-ignored-files
+                  projectile-globally-ignored-files
+                  projectile-globally-ignored-file-suffixes
+                  projectile-globally-ignored-directories)
+              (projectile-grep "hi")))))))
+
+  (describe "rgrep"
+    (before-each
+      (spy-on 'compilation-start))
+    (it "excludes global ignores"
+      (projectile-test-with-sandbox
+       (projectile-test-with-files
+        ("project/"
+         "project/.projectile")
+        (cd "project")
+        (with-current-buffer (find-file-noselect ".projectile" t)
+          (let ((grep-find-template "<X>")
+                (grep-find-ignored-directories '("IG_DIR"))
+                (grep-find-ignored-files '("IG_FILE"))
+                (projectile-globally-ignored-files '("GLOB_IG_FILE"))
+                (projectile-globally-ignored-file-suffixes '("IG_SUF"))
+                (projectile-globally-ignored-directories '("GLOB_IG_DIR")))
+            (projectile-grep "hi")))
+        (expect 'compilation-start :to-have-been-called-with
+                (concat "-type d \\( -path \\*/IG_DIR \\) -prune -o "
+                        "\\! -type d \\( -name IG_FILE -o -name \\*IG_SUF \\) -prune -o "
+                        "\\( -path ./GLOB_IG_DIR -o -path ./GLOB_IG_FILE \\) -prune -o ")
+                'grep-mode))))
+    (it "excludes project ignores"
+      (projectile-test-with-sandbox
+       (projectile-test-with-files
+        ("project/bar/"
+         "project/baz/")
+        (cd "project")
+        (with-temp-file ".projectile" (insert (concat "-/*.txt\n"
+                                                      "-/bar/*.txt\n"
+                                                      "-/baz\n"
+                                                      "-*.txt\n"
+                                                      "-*.text\n"
+                                                      "!/abc.txt\n"
+                                                      "!/bar/abc.txt\n"
+                                                      "!def.txt\n")))
+        (with-temp-file "foo.txt")
+        (with-temp-file "abc.txt")
+        (with-temp-file "bar/foo.txt")
+        (with-temp-file "bar/abc.txt")
+        (with-current-buffer (find-file-noselect ".projectile" t)
+          (let ((grep-find-template "<X>")
+                grep-find-ignored-directories grep-find-ignored-files
+                projectile-globally-ignored-files
+                projectile-globally-ignored-file-suffixes
+                projectile-globally-ignored-directories)
+            (projectile-grep "hi")))
+        (expect 'compilation-start :to-have-been-called-with
+                (concat "\\( -path ./baz -o -path ./foo.txt -o -path ./bar/foo.txt \\) -prune -o "
+                        "\\( "
+                        "\\( -path \\*.txt -o -path \\*.text \\) "
+                        "-a \\! \\( -path ./abc.txt -o -path ./bar/abc.txt -o -path \\*/def.txt \\) "
+                        "\\) -prune -o ")
+                'grep-mode)))))
   (it "grep a git project using default files"
     (require 'vc-git)
     (projectile-test-with-sandbox
@@ -669,74 +770,125 @@ test temp directory"
 
 (describe "projectile-get-other-files"
   (it "returns files with same names but different extensions"
-    (let ((projectile-other-file-alist '(;; handle C/C++ extensions
-                                         ("cpp" . ("h" "hpp" "ipp"))
-                                         ("ipp" . ("h" "hpp" "cpp"))
-                                         ("hpp" . ("h" "ipp" "cpp"))
-                                         ("cxx" . ("hxx" "ixx"))
-                                         ("ixx" . ("cxx" "hxx"))
-                                         ("hxx" . ("ixx" "cxx"))
-                                         ("c" . ("h"))
-                                         ("m" . ("h"))
-                                         ("mm" . ("h"))
-                                         ("h" . ("c" "cpp" "ipp" "hpp" "m" "mm"))
-                                         ("cc" . ("hh"))
-                                         ("hh" . ("cc"))
+    (projectile-test-with-sandbox
+      (projectile-test-with-files-using-custom-project
+          ("src/test1.c"
+           "src/test2.c"
+           "src/test+copying.m"
+           "src/test1.cpp"
+           "src/test2.cpp"
+           "src/Makefile"
+           "src/test.vert"
+           "src/test.frag"
+           "src/same_name.c"
+           "src/some_module/same_name.c"
+           "include1/same_name.h"
+           "include1/test1.h"
+           "include1/test1.h~"
+           "include1/test2.h"
+           "include1/test+copying.h"
+           "include1/test1.hpp"
+           "include2/some_module/same_name.h"
+           "include2/test1.h"
+           "include2/test2.h"
+           "include2/test2.hpp"
+           "src/test1.service.js"
+           "src/test2.service.spec.js"
+           "include1/test1.service.spec.js"
+           "include2/test1.service.spec.js"
+           "include1/test2.js"
+           "include2/test2.js")
+          ()
+        (let ((projectile-other-file-alist '(;; handle C/C++ extensions
+                                             ("cpp" . ("h" "hpp" "ipp"))
+                                             ("ipp" . ("h" "hpp" "cpp"))
+                                             ("hpp" . ("h" "ipp" "cpp"))
+                                             ("cxx" . ("hxx" "ixx"))
+                                             ("ixx" . ("cxx" "hxx"))
+                                             ("hxx" . ("ixx" "cxx"))
+                                             ("c" . ("h"))
+                                             ("m" . ("h"))
+                                             ("mm" . ("h"))
+                                             ("h" . ("c" "cpp" "ipp" "hpp" "m" "mm"))
+                                             ("cc" . ("hh"))
+                                             ("hh" . ("cc"))
 
-                                         ;; vertex shader and fragment shader extensions in glsl
-                                         ("vert" . ("frag"))
-                                         ("frag" . ("vert"))
+                                             ;; vertex shader and fragment shader extensions in glsl
+                                             ("vert" . ("frag"))
+                                             ("frag" . ("vert"))
 
-                                         ;; handle files with no extension
-                                         (nil . ("lock" "gpg"))
-                                         ("lock" . (""))
-                                         ("gpg" . (""))
+                                             ;; handle files with no extension
+                                             (nil . ("lock" "gpg"))
+                                             ("lock" . (""))
+                                             ("gpg" . (""))
 
-                                         ;; handle files with nested extensions
-                                         ("service.js" . ("service.spec.js"))
-                                         ("js" . ("js"))))
-          (source-tree '("src/test1.c"
-                       "src/test2.c"
-                       "src/test+copying.m"
-                       "src/test1.cpp"
-                       "src/test2.cpp"
-                       "src/Makefile"
-                       "src/test.vert"
-                       "src/test.frag"
-                       "src/same_name.c"
-                       "src/some_module/same_name.c"
-                       "include1/same_name.h"
-                       "include1/test1.h"
-                       "include1/test1.h~"
-                       "include1/test2.h"
-                       "include1/test+copying.h"
-                       "include1/test1.hpp"
-                       "include2/some_module/same_name.h"
-                       "include2/test1.h"
-                       "include2/test2.h"
-                       "include2/test2.hpp"
+                                             ;; handle files with nested extensions
+                                             ("service.js" . ("service.spec.js"))
+                                             ("js" . ("js")))))
+          (expect (projectile-get-other-files "src/test1.c") :to-equal '("include1/test1.h" "include2/test1.h"))
+          (expect (projectile-get-other-files "src/test1.cpp") :to-equal '("include1/test1.h" "include2/test1.h" "include1/test1.hpp"))
+          (expect (projectile-get-other-files "test2.c") :to-equal '("include1/test2.h" "include2/test2.h"))
+          (expect (projectile-get-other-files "test2.cpp") :to-equal '("include1/test2.h" "include2/test2.h" "include2/test2.hpp"))
+          (expect (projectile-get-other-files "test1.h") :to-equal '("src/test1.c" "src/test1.cpp" "include1/test1.hpp"))
+          (expect (projectile-get-other-files "test2.h") :to-equal '("src/test2.c" "src/test2.cpp" "include2/test2.hpp"))
+          (expect (projectile-get-other-files "include1/test1.h" t) :to-equal '("src/test1.c" "src/test1.cpp" "include1/test1.hpp"))
+          (expect (projectile-get-other-files "Makefile.lock") :to-equal '("src/Makefile"))
+          (expect (projectile-get-other-files "include2/some_module/same_name.h") :to-equal '("src/some_module/same_name.c" "src/same_name.c"))
+          ;; nested extensions
+          (expect (projectile-get-other-files "src/test1.service.js") :to-equal '("include1/test1.service.spec.js" "include2/test1.service.spec.js"))
+          ;; fallback to outer extensions if no rule for nested extension defined
+          (expect (projectile-get-other-files "src/test2.service.spec.js") :to-equal '("include1/test2.js" "include2/test2.js"))
+          (expect (projectile-get-other-files "src/test+copying.m") :to-equal '("include1/test+copying.h"))))))
 
-                       "src/test1.service.js"
-                       "src/test2.service.spec.js"
-                       "include1/test1.service.spec.js"
-                       "include2/test1.service.spec.js"
-                       "include1/test2.js"
-                       "include2/test2.js")))
+  (it "returns files based on the paths returned by :related-files-fn option"
+    (projectile-test-with-sandbox
+      (projectile-test-with-files-using-custom-project
+          ("src/test1.cpp"
+           "src/test1.def"
+           "src/test2.def"
+           "src/test2.cpp"
+           "src/test2.h"
+           "src/test3.cpp"
+           "src/test3.h")
+          (:related-files-fn (lambda (file)
+                           (cond ((equal file "src/test1.def") '(:other "src/test1.cpp"))
+                                 ((equal file "src/test2.def") '(:other ("src/test2.cpp" "src/test2.h" "src/test4.h")))
+                                 ((equal file "src/test3.cpp") '(:other nil)))))
+        (expect (projectile-get-other-files "src/test1.def") :to-equal '("src/test1.cpp"))
+        (expect (projectile-get-other-files "src/test2.def") :to-equal '("src/test2.cpp" "src/test2.h"))
+        ;; Make sure extension based mechanism is still working
+        (expect (projectile-get-other-files "src/test2.cpp") :to-equal '("src/test2.h"))
+        ;; Make sure that related-files-fn option has priority over existing mechanism
+        (expect (projectile-get-other-files "src/test3.cpp") :to-equal nil))))
 
-      (expect (projectile-get-other-files "src/test1.c" source-tree) :to-equal '("include1/test1.h" "include2/test1.h"))
-      (expect (projectile-get-other-files "src/test1.cpp" source-tree) :to-equal '("include1/test1.h" "include2/test1.h" "include1/test1.hpp"))
-      (expect (projectile-get-other-files "test2.c" source-tree) :to-equal '("include1/test2.h" "include2/test2.h"))
-      (expect (projectile-get-other-files "test2.cpp" source-tree) :to-equal '("include1/test2.h" "include2/test2.h" "include2/test2.hpp"))
-      (expect (projectile-get-other-files "test1.h" source-tree) :to-equal '("src/test1.c" "src/test1.cpp" "include1/test1.hpp"))
-      (expect (projectile-get-other-files "test2.h" source-tree) :to-equal '("src/test2.c" "src/test2.cpp" "include2/test2.hpp"))
-      (expect (projectile-get-other-files "include1/test1.h" source-tree t) :to-equal '("src/test1.c" "src/test1.cpp" "include1/test1.hpp"))
-      (expect (projectile-get-other-files "Makefile.lock" source-tree) :to-equal '("src/Makefile"))
-      (expect (projectile-get-other-files "include2/some_module/same_name.h" source-tree) :to-equal '("src/some_module/same_name.c" "src/same_name.c"))
-      ;; nested extensions
-      (expect (projectile-get-other-files "src/test1.service.js" source-tree) :to-equal '("include1/test1.service.spec.js" "include2/test1.service.spec.js"))
-      ;; fallback to outer extensions if no rule for nested extension defined
-      (expect (projectile-get-other-files "src/test2.service.spec.js" source-tree) :to-equal '("include1/test2.js" "include2/test2.js"))
-      (expect (projectile-get-other-files "src/test+copying.m" source-tree) :to-equal '("include1/test+copying.h")))))
+  (it "returns files based on the predicate returned by :related-files-fn option"
+    (projectile-test-with-sandbox
+      (projectile-test-with-files-using-custom-project
+          ("src/test1.cpp"
+           "src/test1.def"
+           "src/test2.def"
+           "src/test2.cpp"
+           "src/test2.h"
+           "src/test3.cpp"
+           "src/test3.h")
+          (:related-files-fn
+           (lambda (file)
+             (cond ((equal file "src/test1.def")
+                    (list :other (lambda (other-file)
+                                   (equal other-file "src/test1.cpp"))))
+                   ((equal file "src/test2.def")
+                    (list :other (lambda (other-file)
+                                   (or (equal other-file "src/test2.cpp")
+                                       (equal other-file "src/test2.h")))))
+                   ((equal file "src/test3.cpp")
+                    (list :other (lambda (other-file) nil))))))
+
+        (expect (projectile-get-other-files "src/test1.def") :to-equal '("src/test1.cpp"))
+        (expect (projectile-get-other-files "src/test2.def") :to-equal '("src/test2.cpp" "src/test2.h"))
+        ;; Make sure extension based mechanism is still working
+        (expect (projectile-get-other-files "src/test2.cpp") :to-equal '("src/test2.h"))
+        ;; Make sure that related-files-fn option has priority over existing mechanism
+        (expect (projectile-get-other-files "src/test3.cpp") :to-equal nil)))))
 
 (describe "projectile-compilation-dir"
   (it "returns the compilation directory for a project"
@@ -812,60 +964,219 @@ test temp directory"
     (expect (projectile-dirname-matching-count "src/weed/sea.c" "src/food/sea.c") :to-equal 0)
     (expect (projectile-dirname-matching-count "test/demo-test.el" "demo.el") :to-equal 0)))
 
-(describe "projectile-find-matching-test"
+(describe "projectile--find-matching-test"
   (it "finds matching test or file"
     (projectile-test-with-sandbox
-     (projectile-test-with-files
-      ("project/app/models/weed/"
-       "project/app/models/food/"
-       "project/spec/models/weed/"
-       "project/spec/models/food/"
-       "project/app/models/weed/sea.rb"
-       "project/app/models/food/sea.rb"
-       "project/spec/models/weed/sea_spec.rb"
-       "project/spec/models/food/sea_spec.rb")
-      (let ((projectile-indexing-method 'native))
-        (spy-on 'projectile-project-type :and-return-value 'rails-rspec)
-        (spy-on 'projectile-project-root :and-return-value (file-truename (expand-file-name "project/")))
-        (expect (projectile-find-matching-test "app/models/food/sea.rb") :to-equal "spec/models/food/sea_spec.rb")
-        (expect (projectile-find-matching-file "spec/models/food/sea_spec.rb") :to-equal "app/models/food/sea.rb")))))
-  (it "finds matching test or file in a custom project"
+      (projectile-test-with-files-using-custom-project
+          ("app/models/weed/sea.rb"
+           "app/models/food/sea.rb"
+           "spec/models/weed/sea_spec.rb"
+           "spec/models/food/sea_spec.rb")
+          (:test-suffix "_spec")
+        (expect (projectile--find-matching-test "app/models/food/sea.rb") :to-equal '("spec/models/food/sea_spec.rb"))
+        (expect (projectile--find-matching-file "spec/models/food/sea_spec.rb") :to-equal '("app/models/food/sea.rb")))))
+
+  (it "finds matching test or file with dirs"
     (projectile-test-with-sandbox
-     (projectile-test-with-files
-      ("project/src/foo/"
-       "project/src/bar/"
-       "project/test/foo/"
-       "project/test/bar/"
-       "project/src/foo/foo.service.js"
-       "project/src/bar/bar.service.js"
-       "project/test/foo/foo.service.spec.js"
-       "project/test/bar/bar.service.spec.js")
-      (let ((projectile-indexing-method 'native))
-        (projectile-register-project-type 'npm-project '("somefile") :test-suffix ".spec")
-        (spy-on 'projectile-project-type :and-return-value 'npm-project)
-        (spy-on 'projectile-project-root :and-return-value (file-truename (expand-file-name "project/")))
-        (expect (projectile-find-matching-test "src/foo/foo.service.js") :to-equal "test/foo/foo.service.spec.js")
-        (expect (projectile-find-matching-file "test/bar/bar.service.spec.js") :to-equal "src/bar/bar.service.js")))))
-  (it "finds matching test or file in a custom project with dirs"
+      (projectile-test-with-files-using-custom-project
+          ("source/foo/foo.service.js"
+           "source/bar/bar.service.js"
+           "spec/foo/foo.service.spec.js"
+           "spec/bar/bar.service.spec.js")
+          (:test-suffix ".spec" :test-dir "spec/" :src-dir "source/")
+        (expect (projectile--find-matching-test "source/foo/foo.service.js") :to-equal '("spec/foo/foo.service.spec.js"))
+        (expect (projectile--find-matching-file "spec/bar/bar.service.spec.js") :to-equal '("source/bar/bar.service.js")))))
+
+  (it "finds matching test or file based on the paths returned by :related-files-fn option"
+    (defun -my/related-files(file)
+      (if (string-match (rx (group (or "src" "test")) (group "/" (1+ anything) ".cpp")) file)
+          (if (equal (match-string 1 file ) "test")
+              (list :impl (concat "src" (match-string 2 file)))
+            (list :test (concat "test" (match-string 2 file))))))
     (projectile-test-with-sandbox
-     (projectile-test-with-files
-      ("project/source/foo/"
-       "project/source/bar/"
-       "project/spec/foo/"
-       "project/spec/bar/"
-       "project/source/foo/foo.service.js"
-       "project/source/bar/bar.service.js"
-       "project/spec/foo/foo.service.spec.js"
-       "project/spec/bar/bar.service.spec.js")
-      (let ((projectile-indexing-method 'native))
-        (projectile-register-project-type 'npm-project '("somefile")
-                                          :test-suffix ".spec"
-                                          :test-dir "spec/"
-                                          :src-dir "source/")
-        (spy-on 'projectile-project-type :and-return-value 'npm-project)
-        (spy-on 'projectile-project-root :and-return-value (file-truename (expand-file-name "project/")))
-        (expect (projectile-find-matching-test "source/foo/foo.service.js") :to-equal "spec/foo/foo.service.spec.js")
-        (expect (projectile-find-matching-file "spec/bar/bar.service.spec.js") :to-equal "source/bar/bar.service.js"))))))
+      (projectile-test-with-files-using-custom-project
+          ("src/Foo.cpp"
+           "src/Bar.cpp"
+           "src/Baz.py"
+           "test/Bar.cpp"
+           "test/Foo.cpp"
+           "other/Test_Baz.py")
+          (:related-files-fn #'-my/related-files :test-prefix "Test_")
+        (expect (projectile-test-file-p "test/Foo.cpp") :to-equal t)
+        (expect (projectile-test-file-p "src/Foo.cpp") :to-equal nil)
+        (expect (projectile--find-matching-test "src/Foo.cpp") :to-equal '("test/Foo.cpp"))
+        (expect (projectile--find-matching-test "src/Foo2.cpp") :to-equal nil)
+        (expect (projectile--find-matching-file "test/Foo.cpp") :to-equal '("src/Foo.cpp"))
+        (expect (projectile--find-matching-file "test/Foo2.cpp") :to-equal nil)
+        ;; Make sure that existing mechanism(:test-prefix) still works
+        (expect (projectile-test-file-p "other/Test_Baz.py") :to-equal t)
+        (expect (projectile-test-file-p "other/Baz.py") :to-equal nil)
+        (expect (projectile--find-matching-file "other/Test_Baz.py") :to-equal '("src/Baz.py"))
+        (expect (projectile--find-matching-test "src/Baz.py") :to-equal '("other/Test_Baz.py")))))
+
+  (it "finds matching test or file by the predicate returned by :related-files-fn option"
+    (defun -my/related-files(file)
+      (cond ((equal file "src/Foo.cpp")
+             (list :test (lambda (other-file)
+                      (equal other-file "test/Foo.cpp"))))
+            ((equal file "test/Foo.cpp")
+             (list :impl (lambda (other-file)
+                      (equal other-file "src/Foo.cpp"))))))
+    (projectile-test-with-sandbox
+      (projectile-test-with-files-using-custom-project
+          ("src/Foo.cpp"
+           "src/Bar.cpp"
+           "test/Bar.cpp"
+           "test/Foo.cpp")
+          (:related-files-fn #'-my/related-files)
+        (expect (projectile-test-file-p "test/Foo.cpp") :to-equal t)
+        (expect (projectile-test-file-p "src/Foo.cpp") :to-equal nil)
+        (expect (projectile--find-matching-test "src/Foo.cpp") :to-equal '("test/Foo.cpp"))
+        (expect (projectile--find-matching-test "src/Foo.cpp") :to-equal '("test/Foo.cpp"))
+        (expect (projectile--find-matching-file "test/Foo.cpp") :to-equal '("src/Foo.cpp"))))))
+
+(describe "projectile--related-files"
+  (it "returns related files for the given file"
+    (defun -my/related-files(file)
+      (cond ((equal file "src/Foo.c")
+             (list :test "src/TestFoo.c" :doc "doc/Foo.txt"))
+            ((equal file "src/TestFoo.c")
+             (list :impl (lambda (other-file)
+                           (equal other-file "src/Foo.c"))))))
+    (projectile-test-with-sandbox
+      (projectile-test-with-files-using-custom-project
+          ("src/Foo.c"
+           "src/TestFoo.c"
+           "doc/Foo.txt")
+          (:related-files-fn #'-my/related-files)
+        (expect (projectile--related-files-kinds "src/Foo.c") :to-equal '(:test :doc))
+        (expect (projectile--related-files-kinds "src/TestFoo.c") :to-equal '(:impl))
+        (expect (projectile--related-files "src/TestFoo.c" :impl) :to-equal '("src/Foo.c"))
+        (expect (projectile--related-files "src/Foo.c" :doc) :to-equal '("doc/Foo.txt"))
+        ;; Support abspath
+        (expect (projectile--related-files-kinds (concat (projectile-project-root) "src/Foo.c")) :to-equal '(:test :doc))
+        (expect (projectile--related-files (concat (projectile-project-root) "src/Foo.c") :doc) :to-equal '("doc/Foo.txt"))))))
+
+(describe "projectile--merge-related-files-fns"
+  (it "returns a new function which returns the merged plist from each fn"
+    (defun -first-fn(file)
+      (list :foo "file1"))
+    (defun -second-fn(file)
+      (list :foo (list "file2" "file3")))
+    (defun -third-fn(file)
+      (list :bar "file4"))
+    (let ((fn (projectile--merge-related-files-fns '(-first-fn -second-fn))))
+      (expect (funcall fn "something") :to-equal '(:foo ("file1" "file2" "file3"))))
+    (let ((fn (projectile--merge-related-files-fns '(-first-fn -third-fn))))
+      (expect (funcall fn "something") :to-equal '(:foo ("file1") :bar ("file4"))))))
+
+(describe "projectile-related-files-fn-groups"
+  (it "generate related files fn which relates members of each group as a specified kind"
+    (let ((fn (projectile-related-files-fn-groups :foo '(("a.cpp" "req/a.txt" "doc/a.uml")
+                                                         ("b.cpp" "req/b.txt")))))
+      (expect (funcall fn "a.cpp") :to-equal '(:foo ("req/a.txt" "doc/a.uml")))
+      (expect (funcall fn "req/a.txt") :to-equal '(:foo ("a.cpp" "doc/a.uml")))
+      (expect (funcall fn "b.cpp") :to-equal '(:foo ("req/b.txt")))
+      (expect (funcall fn "c.cpp") :to-equal nil))))
+
+(describe "projectile-related-files-fn-extensions"
+  (it "generate related files fn which relates files with the given extnsions"
+    (let* ((fn (projectile-related-files-fn-extensions :foo '("cpp" "h" "hpp")))
+           (plist (funcall fn "a.cpp"))
+           (predicate (plist-get plist :foo)))
+      (expect plist :to-contain :foo)
+      (expect (funcall predicate "a.h") :to-equal t)
+      (expect (funcall predicate "a.hpp") :to-equal t)
+      (expect (funcall predicate "b.cpp") :to-equal nil)
+      (expect (funcall predicate "a.cpp") :to-equal nil))))
+
+(describe "projectile-related-files-fn-tests-with-prefix"
+  (it "generate related files fn which relates tests and impl based on extension and prefix"
+    (let ((fn (projectile-related-files-fn-test-with-prefix "py" "test_")))
+      (let* ((plist (funcall fn "foo/a.py"))
+            (predicate (plist-get plist :test)))
+        (expect plist :to-contain :test)
+        (expect (funcall predicate "bar/test_a.py") :to-equal t)
+        (expect (funcall predicate "bar/test_a.cpp") :to-equal nil))
+      (let* ((plist (funcall fn "foo/test_a.py"))
+             (predicate (plist-get plist :impl)))
+        (expect plist :to-contain :impl)
+        (expect (funcall predicate "bar/a.py") :to-equal t)
+        (expect (funcall predicate "bar/a.cpp") :to-equal nil)
+        (expect (funcall predicate "bar/test_a.cpp") :to-equal nil)))))
+
+(describe "projectile-related-files-fn-tests-with-suffix"
+  (it "generate related files fn which relates tests and impl based on extension and suffix"
+    (let ((fn (projectile-related-files-fn-test-with-suffix "py" "-test")))
+      (let* ((plist (funcall fn "foo/a.py"))
+            (predicate (plist-get plist :test)))
+        (expect plist :to-contain :test)
+        (expect (funcall predicate "bar/a-test.py") :to-equal t)
+        (expect (funcall predicate "bar/a-test.cpp") :to-equal nil))
+      (let* ((plist (funcall fn "foo/a-test.py"))
+             (predicate (plist-get plist :impl)))
+        (expect plist :to-contain :impl)
+        (expect (funcall predicate "bar/a.py") :to-equal t)
+        (expect (funcall predicate "bar/a.cpp") :to-equal nil)
+        (expect (funcall predicate "bar/a-test.cpp") :to-equal nil)))))
+
+(describe "projectile--related-files-plist-by-kind"
+  (defun -sample-predicate (other-file)
+    (equal other-file "src/foo.c"))
+  (defun -sample-predicate2 (other-file)
+    (equal other-file "src/bar.c"))
+  (describe "when :related-files-fn returns paths"
+    (it "returns a plist containing :paths only with the existing files on file system without duplication"
+      (projectile-test-with-sandbox
+        (projectile-test-with-files-using-custom-project
+            ("src/foo.c")
+            (:related-files-fn (lambda (_)
+                                 (list :foo '("src/foo.c" "src/bar.c" "src/foo.c"))))
+          (expect (projectile--related-files-plist-by-kind "something" :foo)
+                  :to-equal '(:paths ("src/foo.c")))))))
+  (describe "when :related-files-fn returns one predicate"
+    (it "returns a plist containing :predicate with the same predicate"
+      (projectile-test-with-sandbox
+        (projectile-test-with-files-using-custom-project
+            ("src/foo.c")  ; Contents does not matter
+            (:related-files-fn (lambda (_)
+                                 (list :foo '-sample-predicate)))
+          (expect (projectile--related-files-plist-by-kind "something" :foo)
+                  :to-equal '(:predicate -sample-predicate))))))
+  (describe "when :related-files-fn returns multiple predicates"
+    (it "returns a plist containing :predicate with a merging predicate"
+      (projectile-test-with-sandbox
+        (projectile-test-with-files-using-custom-project
+            ("src/foo.c")  ; Contents does not matter
+            (:related-files-fn (lambda (_)
+                                 (list :foo (list '-sample-predicate '-sample-predicate2))))
+          (let* ((plist (projectile--related-files-plist-by-kind "something" :foo))
+                 (predicate (plist-get plist :predicate)))
+            (expect plist :to-contain :predicate)
+            (expect (funcall predicate "src/foo.c") :to-equal t)
+            (expect (funcall predicate "src/bar.c") :to-equal t))))))
+  (describe "when :related-files-fn returns both paths and predicates"
+    (it "returns a plist containing both :paths and :predicates"
+      (projectile-test-with-sandbox
+        (projectile-test-with-files-using-custom-project
+            ("src/foo.c")
+            (:related-files-fn (lambda (_)
+                                 (list :foo '("src/foo.c" -sample-predicate))))
+          (expect (projectile--related-files-plist-by-kind "something" :foo)
+                  :to-equal '(:paths ("src/foo.c") :predicate -sample-predicate))))))
+  (describe "when :related-files-fn is a list of functions"
+    (it "returns a plist containing the merged results"
+      (defun -sample-fn(file)
+        (list :foo "src/foo.c"))
+      (defun -sample-fn2(file)
+        (list :foo '-sample-predicate))
+      (projectile-test-with-sandbox
+        (projectile-test-with-files-using-custom-project
+            ("src/foo.c"
+             "src/bar.c")
+            (:related-files-fn (list '-sample-fn '-sample-fn2))
+          (expect (projectile--related-files-plist-by-kind "something" :foo)
+                  :to-equal '(:paths ("src/foo.c") :predicate -sample-predicate)))))))
 
 (describe "projectile-get-all-sub-projects"
   (it "excludes out-of-project submodules"
